class BrotherQLError(Exception):
    pass

class BrotherQLUnsupportedCmd(BrotherQLError):
    pass

class BrotherQLUnknownModel(BrotherQLError):
    pass

class BrotherQLRasterError(BrotherQLError):
    pass



from .converter import BrotherLabelConverter
from .devices import BrotherDeviceManager

class BrotherLabel(object):
    def __init__(self, device, strict=False):
        self.strict = strict

        self.converter = BrotherLabelConverter()
        self.devices = BrotherDeviceManager()

        if isinstance(device, str):
            self.device = self.devices[device]
        else:
            self.device = device

    def convert(self, type, images,  **kwargs):
        return self.converter.convert(self.device, type, images, **kwargs)




from builtins import bytes

import logging

logger = logging.getLogger(__name__)

def list_available_devices():
    """ List all available devices for the respective backend """
    # returns a list of dictionaries with the keys 'identifier' and 'instance':
    # [ {'identifier': '/dev/usb/lp0', 'instance': os.open('/dev/usb/lp0', os.O_RDWR)}, ]
    raise NotImplementedError()


class BrotherQLBackendGeneric(object):

    def __init__(self, device_specifier):
        """
        device_specifier can be either a string or an instance
        of the required class type.
        """
        self.write_dev = None
        self.read_dev  = None
        raise NotImplementedError()

    def _write(self, data):
        self.write_dev.write(data)

    def _read(self, length=32):
        return bytes(self.read_dev.read(length))

    def write(self, data):
        logger.debug('Writing %d bytes.', len(data))
        self._write(data)

    def read(self, length=32):
        try:
            ret_bytes = self._read(length)
            if ret_bytes: logger.debug('Read %d bytes.', len(ret_bytes))
            return ret_bytes
        except Exception as e:
            logger.debug('Error reading... %s', e)
            raise

    def dispose(self):
        try:
            self._dispose()
        except:
            pass

    def _dispose(self):
        raise NotImplementedError()

    def __del__(self):
        self.dispose()



#!/usr/bin/env python

"""
Backend to support Brother QL-series printers via network.
Works cross-platform.
"""

from __future__ import unicode_literals
from builtins import str

import socket, os, time, select

from .generic import BrotherQLBackendGeneric

def list_available_devices():
    """
    List all available devices for the network backend

    returns: devices: a list of dictionaries with the keys 'identifier' and 'instance': \
        [ {'identifier': 'tcp://hostname[:port]', 'instance': None}, ] \
        Instance is set to None because we don't want to connect to the device here yet.
    """

    # We need some snmp request sent to 255.255.255.255 here
    raise NotImplementedError()
    return [{'identifier': 'tcp://' + path, 'instance': None} for path in paths]

class BrotherQLBackendNetwork(BrotherQLBackendGeneric):
    """
    BrotherQL backend using the Linux Kernel USB Printer Device Handles
    """

    def __init__(self, device_specifier):
        """
        device_specifier: string or os.open(): identifier in the \
            format file:///dev/usb/lp0 or os.open() raw device handle.
        """

        self.read_timeout = 0.01
        # strategy : try_twice, select or socket_timeout
        self.strategy = 'socket_timeout'
        if isinstance(device_specifier, str):
            if device_specifier.startswith('tcp://'):
                device_specifier = device_specifier[6:]
            host, _, port = device_specifier.partition(':')
            if port:
                port = int(port)
            else:
                port = 9100
            #try:
            self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.s.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            self.s.connect((host, port))
            #except OSError as e:
            #    raise ValueError('Could not connect to the device.')
            if self.strategy == 'socket_timeout':
                self.s.settimeout(self.read_timeout)
            elif self.strategy == 'try_twice':
                self.s.settimeout(self.read_timeout)
            else:
                self.s.settimeout(0)

        elif isinstance(device_specifier, int):
            self.dev = device_specifier
        else:
            raise NotImplementedError('Currently the printer can be specified either via an appropriate string or via an os.open() handle.')

    def _write(self, data):
        self.s.settimeout(10)
        self.s.sendall(data)
        self.s.settimeout(self.read_timeout)

    def _read(self, length=32):
        if self.strategy in ('socket_timeout', 'try_twice'):
            if self.strategy == 'socket_timeout':
                tries = 1
            if self.strategy == 'try_twice':
                tries = 2
            for i in range(tries):
                try:
                    data = self.s.recv(length)
                    return data
                except socket.timeout:
                    pass
            return b''
        elif self.strategy == 'select':
            data = b''
            start = time.time()
            while (not data) and (time.time() - start < self.read_timeout):
                result, _, _ = select.select([self.s], [], [], 0)
                if self.s in result:
                    data += self.s.recv(length)
                if data: break
                time.sleep(0.001)
            return data
        else:
            raise NotImplementedError('Unknown strategy')

    def _dispose(self):
        self.s.shutdown(socket.SHUT_RDWR)
        self.s.close()



"""
This module contains the implementation of the raster language
of the Brother QL-series label printers according to their
documentation and to reverse engineering efforts.

The central piece of code in this module is the class
:py:class:`BrotherQLRaster`.
"""

from builtins import bytes

import struct
import logging

import packbits
from PIL import Image
from io import BytesIO
import io

from . import BrotherQLError, BrotherQLUnsupportedCmd, BrotherQLUnknownModel, BrotherQLRasterError

logger = logging.getLogger(__name__)

class BrotherLabelRaster(object):
    """
    This class facilitates the creation of a complete set
    of raster instructions by adding them one after the other
    using the methods of the class. Each method call is adding
    instructions to the member variable :py:attr:`data`.

    Instatiate the class by providing the printer
    model as argument.

    :param str model: Choose from the list of available models.

    :ivar bytes data: The resulting bytecode with all instructions.
    :ivar bool exception_on_warning: If set to True, an exception is raised if trying to add instruction which are not supported on the selected model. If set to False, the instruction is simply ignored and a warning sent to logging/stderr.
    """

    def __init__(self, device):
        self.device = device
        
        self.data = b''
        self._pquality = True
        self.page_number = 0
        self.cut_at_end = True
        self.dpi_600 = False
        self.two_color_printing = False
        self._compression = False
        self.exception_on_warning = False

    def _warn(self, problem, kind=BrotherQLRasterError):
        """
        Logs the warning message `problem` or raises a
        `BrotherQLRasterError` exception (changeable via `kind`)
        if `self.exception_on_warning` is set to True.

        :raises BrotherQLRasterError: Or other exception \
            set via the `kind` keyword argument.
        """
        if self.exception_on_warning:
            raise kind(problem)
        else:
            logger.warning(problem)

    def _unsupported(self, problem):
        """
        Raises BrotherQLUnsupportedCmd if
        exception_on_warning is set to True.
        Issues a logger warning otherwise.

        :raises BrotherQLUnsupportedCmd:
        """
        self._warn(problem, kind=BrotherQLUnsupportedCmd)

    def add_initialize(self):
        self.page_number = 0
        self.data += b'\x1B\x40' # ESC @

    def add_status_information(self):
        """ Status Information Request """
        self.data += b'\x1B\x69\x53' # ESC i S

    def add_switch_mode(self):
        """
        Switch dynamic command mode
        Switch to the raster mode on the printers that support
        the mode change (others are in raster mode already).
        """
        if not self.device.mode_setting:
            self._unsupported("Trying to switch the operating mode on a printer that doesn't support the command.")
            return

        self.data += b'\x1B\x69\x61\x01' # ESC i a

    def add_invalidate(self):
        """ clear command buffer """
        self.data += b'\x00' * self.device.num_invalidate_bytes

    @property
    def mtype(self): return self._mtype

    @property
    def mwidth(self): return self._mwidth

    @property
    def mlength(self): return self._mlength

    @property
    def pquality(self): return self._pquality

    @mtype.setter
    def mtype(self, value):
        self._mtype = bytes([value & 0xFF])

    @mwidth.setter
    def mwidth(self, value):
        self._mwidth = bytes([value & 0xFF])

    @mlength.setter
    def mlength(self, value):
        self._mlength = bytes([value & 0xFF])

    @pquality.setter
    def pquality(self, value):
        self._pquality = bool(value)

    def add_media_and_quality(self, rnumber):
        self.data += b'\x1B\x69\x7A' # ESC i z
        valid_flags = 0x80
        valid_flags |= (self._mtype is not None) << 1
        valid_flags |= (self._mwidth is not None) << 2
        valid_flags |= (self._mlength is not None) << 3
        valid_flags |= self._pquality << 6
        self.data += bytes([valid_flags])
        vals = [self._mtype, self._mwidth, self._mlength]
        self.data += b''.join(b'\x00' if val is None else val for val in vals)
        self.data += struct.pack('<L', rnumber)
        self.data += bytes([0 if self.page_number == 0 else 1])
        self.data += b'\x00'
        # INFO:  media/quality (1B 69 7A) --> found! (payload: 8E 0A 3E 00 D2 00 00 00 00 00)

    def add_autocut(self, autocut = False):
        if not self.device.cutting:
            self._unsupported("Trying to call add_autocut with a printer that doesn't support it")
            return
        
        self.data += b'\x1B\x69\x4D' # ESC i M
        self.data += bytes([autocut << 6])

    def add_cut_every(self, n=1):
        if not self.device.cutting:
            self._unsupported("Trying to call add_cut_every with a printer that doesn't support it")
            return
    
        self.data += b'\x1B\x69\x41' # ESC i A
        self.data += bytes([n & 0xFF])

    def add_expanded_mode(self):
        if not self.device.expanded_mode:
            self._unsupported("Trying to set expanded mode (dpi/cutting at end) on a printer that doesn't support it")
            return
        
        if self.two_color_printing and not self.device.two_color:
            self._unsupported("Trying to set two_color_printing in expanded mode on a printer that doesn't support it.")
            return
        
        self.data += b'\x1B\x69\x4B' # ESC i K
        flags = 0x00
        flags |= self.cut_at_end << 3
        flags |= self.dpi_600 << 6
        flags |= self.two_color_printing << 0
        self.data += bytes([flags])

    def add_margins(self, dots=0x23):
        self.data += b'\x1B\x69\x64' # ESC i d
        self.data += struct.pack('<H', dots)

    def add_compression(self, compression=True):
        """
        Add an instruction enabling or disabling compression for the transmitted raster image lines.
        Not all models support compression. If the specific model doesn't support it but this method
        is called trying to enable it, either a warning is set or an exception is raised depending on
        the value of :py:attr:`exception_on_warning`

        :param bool compression: Whether compression should be on or off
        """
        if not self.device.compression:
            self._unsupported("Trying to set compression on a printer that doesn't support it")
            return

        self._compression = compression
        self.data += b'\x4D' # M
        self.data += bytes([compression << 1])

    def get_pixel_width(self):
        return self.device.number_bytes_per_row * 8

    def add_raster_data(self, image, second_image=None):
        """
        Add the image data to the instructions.
        The provided image has to be binary (every pixel
        is either black or white).

        :param PIL.Image.Image image: The image to be converted and added to the raster instructions
        :param PIL.Image.Image second_image: A second image with a separate color layer (red layer for the QL-800 series)
        """
        logger.debug("raster_image_size: {0}x{1}".format(*image.size))
        if image.size[0] != self.get_pixel_width():
            fmt = 'Wrong pixel width: {}, expected {}'
            raise BrotherQLRasterError(fmt.format(image.size[0], self.get_pixel_width()))
        images = [image]
        if second_image:
            if image.size != second_image.size:
                fmt = "First and second image don't have the same dimesions: {} vs {}."
                raise BrotherQLRasterError(fmt.format(image.size, second_image.size))
            images.append(second_image)
        frames = []
        for image in images:
            image = image.transpose(Image.FLIP_LEFT_RIGHT)
            image = image.convert("1")
            frames.append(bytes(image.tobytes(encoder_name='raw')))
        frame_len = len(frames[0])
        row_len = images[0].size[0]//8
        start = 0
        file_str = BytesIO()
        while start + row_len <= frame_len:
            for i, frame in enumerate(frames):
                row = frame[start:start+row_len]
                if self._compression:
                    row = packbits.encode(row)
                translen = len(row) # number of bytes to be transmitted
                if self.device.identifier.startswith('PT'):
                    file_str.write(b'\x47')
                    file_str.write(bytes([translen%256, translen//256]))
                else:
                    if second_image:
                        file_str.write(b'\x77\x01' if i == 0 else b'\x77\x02')
                    else:
                        file_str.write(b'\x67\x00')
                    file_str.write(bytes([translen]))
                file_str.write(row)
            start += row_len
        self.data += file_str.getvalue()

    def add_print(self, last_page=True):
        if last_page:
            self.data += b'\x1A' # 0x1A = ^Z = SUB; here: EOF = End of File
        else:
            self.data += b'\x0C' # 0x0C = FF  = Form Feed



#!/usr/bin/env python

from __future__ import division, unicode_literals
from builtins import str

import logging

from PIL import Image
import PIL.ImageOps, PIL.ImageChops

from . import BrotherQLUnsupportedCmd
from .labels import FormFactor
from .raster import BrotherLabelRaster

logger = logging.getLogger(__name__)

class BrotherLabelConverter(object):
    def convert(self, device, type, images,  **kwargs):
        r"""Converts one or more images to a raster instruction file.

        :param device:
            An instance of the BrotherDevice class
        :type device: :py:class:`brother_label.devices.BrotherDevice`
        :param str type:
            Type of label the printout should be on.
        :param images:
            The images to be converted. They can be filenames or instances of Pillow's Image.
        :type images: list(PIL.Image.Image) or list(str) images
        :param \**kwargs:
            See below

        :Keyword Arguments:
            * **cut** (``bool``) --
              Enable cutting after printing the labels.
            * **dither** (``bool``) --
              Instead of applying a threshold to the pixel values, approximate grey tones with dithering.
            * **compress**
            * **red**
            * **rotate**
            * **dpi_600**
            * **hq**
            * **threshold**
        """
        label = device.labels_by_id[type]
        raster = BrotherLabelRaster(device)

        right_margin_dots = label.offset_r + device.additional_offset_r
        device_pixel_width = raster.get_pixel_width()

        cut = kwargs.get('cut', True)
        dither = kwargs.get('dither', False)
        compress = kwargs.get('compress', False)
        red = kwargs.get('red', False)
        rotate = kwargs.get('rotate', 'auto')
        if rotate != 'auto': rotate = int(rotate)
        dpi_600 = kwargs.get('dpi_600', False)
        hq = kwargs.get('hq', True)
        threshold = kwargs.get('threshold', 70)
        threshold = 100.0 - threshold
        threshold = min(255, max(0, int(threshold/100.0 * 255)))

        if red and not device.two_color:
            raise BrotherQLUnsupportedCmd('Printing in red is not supported with the selected model.')

        try:
            raster.add_switch_mode()
        except BrotherQLUnsupportedCmd:
            pass
        raster.add_invalidate()
        raster.add_initialize()
        try:
            raster.add_switch_mode()
        except BrotherQLUnsupportedCmd:
            pass

        for image in images:
            if isinstance(image, Image.Image):
                im = image
            else:
                try:
                    im = Image.open(image)
                except:
                    raise NotImplementedError("The image argument needs to be an Image() instance, the filename to an image, or a file handle.")

            if im.mode.endswith('A'):
                # place in front of white background and get red of transparency
                bg = Image.new("RGB", im.size, (255,255,255))
                bg.paste(im, im.split()[-1])
                im = bg
            elif im.mode == "P":
                # Convert GIF ("P") to RGB
                im = im.convert("RGB" if red else "L")
            elif im.mode == "L" and red:
                # Convert greyscale to RGB if printing on black/red tape
                im = im.convert("RGB")

            if dpi_600:
                dots_expected = [el*2 for el in label.dots_printable]
            else:
                dots_expected = label.dots_printable

            if label.form_factor in (FormFactor.ENDLESS, FormFactor.PTOUCH_ENDLESS):
                if rotate not in ('auto', 0):
                    im = im.rotate(rotate, expand=True)
                if dpi_600:
                    im = im.resize((im.size[0]//2, im.size[1]))
                if im.size[0] != label.dots_printable[0]:
                    hsize = int((label.dots_printable[0] / im.size[0]) * im.size[1])
                    im = im.resize((label.dots_printable[0], hsize), Image.NEAREST)
                    logger.warning('Need to resize the image...')
                if im.size[0] < device_pixel_width:
                    new_im = Image.new(im.mode, (device_pixel_width, im.size[1]), (255,)*len(im.mode))
                    new_im.paste(im, (device_pixel_width-im.size[0]-right_margin_dots, 0))
                    im = new_im
            elif label.form_factor in (FormFactor.DIE_CUT, FormFactor.ROUND_DIE_CUT):
                if rotate == 'auto':
                    if im.size[0] == dots_expected[1] and im.size[1] == dots_expected[0]:
                        im = im.rotate(90, expand=True)
                elif rotate != 0:
                    im = im.rotate(rotate, expand=True)
                if im.size[0] != dots_expected[0] or im.size[1] != dots_expected[1]:
                    raise ValueError("Bad image dimensions: %s. Expecting: %s." % (im.size, dots_expected))
                if dpi_600:
                    im = im.resize((im.size[0]//2, im.size[1]))
                new_im = Image.new(im.mode, (device_pixel_width, dots_expected[1]), (255,)*len(im.mode))
                new_im.paste(im, (device_pixel_width-im.size[0]-right_margin_dots, 0))
                im = new_im

            if red:
                filter_h = lambda h: 255 if (h <  40 or h > 210) else 0
                filter_s = lambda s: 255 if s > 100 else 0
                filter_v = lambda v: 255 if v >  80 else 0
                red_im = self.hsv_filter(im, filter_h, filter_s, filter_v)
                red_im = red_im.convert("L")
                red_im = PIL.ImageOps.invert(red_im)
                red_im = red_im.point(lambda x: 0 if x < threshold else 255, mode="1")

                filter_h = lambda h: 255
                filter_s = lambda s: 255
                filter_v = lambda v: 255 if v <  80 else 0
                black_im = self.hsv_filter(im, filter_h, filter_s, filter_v)
                black_im = black_im.convert("L")
                black_im = PIL.ImageOps.invert(black_im)
                black_im = black_im.point(lambda x: 0 if x < threshold else 255, mode="1")
                black_im = PIL.ImageChops.subtract(black_im, red_im)
            else:
                im = im.convert("L")
                im = PIL.ImageOps.invert(im)

                if dither:
                    im = im.convert("1", dither=Image.FLOYDSTEINBERG)
                else:
                    im = im.point(lambda x: 0 if x < threshold else 255, mode="1")

            raster.add_status_information()
            tape_size = label.tape_size
            if label.form_factor in (FormFactor.DIE_CUT, FormFactor.ROUND_DIE_CUT):
                raster.mtype = 0x0B
                raster.mwidth = tape_size[0]
                raster.mlength = tape_size[1]
            elif label.form_factor in (FormFactor.ENDLESS, ):
                raster.mtype = 0x0A
                raster.mwidth = tape_size[0]
                raster.mlength = 0
            elif label.form_factor in (FormFactor.PTOUCH_ENDLESS, ):
                raster.mtype = 0x00
                raster.mwidth = tape_size[0]
                raster.mlength = 0
            raster.pquality = int(hq)
            raster.add_media_and_quality(im.size[1])
            try:
                if cut:
                    raster.add_autocut(True)
                    raster.add_cut_every(1)
            except BrotherQLUnsupportedCmd:
                pass
            try:
                raster.dpi_600 = dpi_600
                raster.cut_at_end = cut
                raster.two_color_printing = True if red else False
                raster.add_expanded_mode()
            except BrotherQLUnsupportedCmd:
                pass
            raster.add_margins(label.feed_margin)
            try:
                if compress: raster.add_compression(True)
            except BrotherQLUnsupportedCmd:
                pass
            if red:
                raster.add_raster_data(black_im, red_im)
            else:
                raster.add_raster_data(im)
            raster.add_print()

        return raster.data

    def hsv_filter(self, im, filter_h, filter_s, filter_v, default_col=(255,255,255)):
        """ https://stackoverflow.com/a/22237709/183995 """

        hsv_im = im.convert('HSV')
        H, S, V = 0, 1, 2
        hsv = hsv_im.split()
        mask_h = hsv[H].point(filter_h)
        mask_s = hsv[S].point(filter_s)
        mask_v = hsv[V].point(filter_v)

        Mdat = []
        for h, s, v in zip(mask_h.getdata(), mask_s.getdata(), mask_v.getdata()):
            Mdat.append(255 if (h and s and v) else 0)

        mask = mask_h
        mask.putdata(Mdat)

        filtered_im = Image.new("RGB", im.size, color=default_col)
        filtered_im.paste(im, None, mask)
        return filtered_im



class ElementManager:
    def __init__(self, items):
        self._items = {}

        for item in items:
            self.register(item)
        
    def register(self, item):
        if item.identifier in self._items:
            raise KeyError('Element Already Registered')
        
        self._items[item.identifier] = item

    def deregister(self, item):
        if item not in self._items:
            return
        
        self._items.pop(item.identifier)

    def items(self):
        return self._items.items()
    
    def keys(self):
        return self._items.keys()

    def values(self):
        return self._items.values()
    
    def __getitem__(self, key):
        return self._items[key]
    
    def __contains__(self, key):
        return self._items.__contains__(key)

    def __iter__(self):
        return self._items.__iter__()



from attr import attrs, attrib
from typing import List, Tuple
from enum import IntEnum

from .core import ElementManager

class FormFactor(IntEnum):
    """
    Enumeration representing the form factor of a label.
    The labels for the Brother QL series are supplied either as die-cut (pre-sized), or for more flexibility the
    continuous label tapes offer the ability to vary the label length.
    """
    #: rectangular die-cut labels
    DIE_CUT = 1
    #: endless (continouse) labels
    ENDLESS = 2
    #: round die-cut labels
    ROUND_DIE_CUT = 3
    #: endless P-touch labels
    PTOUCH_ENDLESS = 4

class Color(IntEnum):
    """
    Enumeration representing the colors to be printed on a label. Most labels only support printing black on white.
    Some newer ones can also print in black and red on white.
    """
    #: The label can be printed in black & white.
    BLACK_WHITE = 0
    #: The label can be printed in black, white & red.
    BLACK_RED_WHITE = 1

@attrs
class Label(object):
    """
    This class represents a label. All specifics of a certain label
    and what the rasterizer needs to take care of depending on the
    label choosen, should be contained in this class.
    """
    #: A string identifier given to each label that can be selected. Eg. '29'.
    identifiers = attrib(type=[str])
    #: The tape size of a single label (width, lenght) in mm. For endless labels, the length is 0 by definition.
    tape_size = attrib(type=Tuple[int, int])
    #: The type of label
    form_factor = attrib(type=FormFactor)
    #: The total area (width, length) of the label in dots (@300dpi).
    dots_total = attrib(type=Tuple[int, int])
    #: The printable area (width, length) of the label in dots (@300dpi).
    dots_printable = attrib(type=Tuple[int, int])
    #: The required offset from the right side of the label in dots to obtain a centered printout.
    offset_r = attrib(type=int)
    #: An additional amount of feeding when printing the label.
    #: This is non-zero for some smaller label sizes and for endless labels.
    feed_margin = attrib(type=int, default=0)
    #: Some labels allow printing in red, most don't.
    color = attrib(type=Color, default=Color.BLACK_WHITE)

    def works_with_model(self, model): # type: bool
        """
        Method to determine if certain label can be printed by the specified printer model.
        """
        if self.restricted_to_models and model not in models: return False
        else: return True

    @property
    def name(self): # type: str
        out = ""

        if self.form_factor in (FormFactor.DIE_CUT,):
            out = '{0}mm x {1}mm die-cut'.format(*self.tape_size)
        elif self.form_factor in (FormFactor.ROUND_DIE_CUT,):
            out = '{0}mm round die-cut'.format(self.tape_size[0])
        else:
            out = '{0}mm endless'.format(self.tape_size[0])

        if self.color == Color.BLACK_RED_WHITE:
            out += ' (black/red/white)'

        return out
    


from attr import attrs, attrib
from typing import Tuple

from .core import ElementManager
from .labels import Color, FormFactor, Label

@attrs
class BrotherDevice(object):
    """
    This class represents a printer model. All specifics of a certain model
    and the opcodes it supports should be contained in this class.
    """
    #: A string identifier given to each model implemented. Eg. 'QL-500'.
    identifier = attrib(type=str)
    #: Minimum and maximum number of rows or 'dots' that can be printed.
    #: Together with the dpi this gives the minimum and maximum length
    #: for continuous tape printing.
    min_max_length_dots = attrib(type=Tuple[int, int])
    #: The minimum and maximum amount of feeding a label
    min_max_feed = attrib(type=Tuple[int, int], default=(35, 1500))
    number_bytes_per_row = attrib(type=int, default=90)
    #: The required additional offset from the right side
    additional_offset_r = attrib(type=int, default=0)
    #: Support for the 'mode setting' opcode
    mode_setting = attrib(type=bool, default=True)
    #: Model has a cutting blade to automatically cut labels
    cutting = attrib(type=bool, default=True)
    #: Model has support for the 'expanded mode' opcode.
    #: (So far, all models that have cutting support do).
    expanded_mode = attrib(type=bool, default=True)
    #: Model has support for compressing the transmitted raster data.
    #: Some models with only USB connectivity don't support compression.
    compression = attrib(type=bool, default=True)
    #: Support for two color printing (black/red/white)
    #: available only on some newer models.
    two_color = attrib(type=bool, default=False)
    #: Number of NULL bytes needed for the invalidate command.
    num_invalidate_bytes = attrib(type=int, default=200)

    def __attrs_post_init__(self) -> None:
        self.labels_by_id = {}

        for label in self.labels:
            for identifier in label.identifiers:
                self.labels_by_id[identifier] = label

    @property
    def labels(self):
        return []

    @property
    def name(self):
        return self.identifier


class BrotherDeviceQL(BrotherDevice):
    @property
    def labels(self):
        return super().labels + [
            # Continuous
            Label(["12", "DK-22214"],                           ( 12,   0), FormFactor.ENDLESS,       ( 142,    0), ( 106,    0),  29 , feed_margin=35),
            Label(["18"],                                       ( 18,   0), FormFactor.ENDLESS,       ( 256,    0), ( 234,    0), 171 , feed_margin=14),
            Label(["29", "DK-22210"],                           ( 29,   0), FormFactor.ENDLESS,       ( 342,    0), ( 306,    0),   6 , feed_margin=35),
            Label(["38", "DK-22225"],                           ( 38,   0), FormFactor.ENDLESS,       ( 449,    0), ( 413,    0),  12 , feed_margin=35),
            Label(["50", "DK-22223"],                           ( 50,   0), FormFactor.ENDLESS,       ( 590,    0), ( 554,    0),  12 , feed_margin=35),
            Label(["54", "DK-N55224"],                          ( 54,   0), FormFactor.ENDLESS,       ( 636,    0), ( 590,    0),   0 , feed_margin=35),
            Label(["62", "DK-22205", "DK-44205", "DK-44605"],   ( 62,   0), FormFactor.ENDLESS,       ( 732,    0), ( 696,    0),  12 , feed_margin=35),
            Label(["62red", "DK-22251"],                        ( 62,   0), FormFactor.ENDLESS,       ( 732,    0), ( 696,    0),  12 , feed_margin=35, color=Color.BLACK_RED_WHITE),
            
            # Die-Cut
            Label(["17x54", "DK-11204"],                        ( 17,  54), FormFactor.DIE_CUT,       ( 201,  636), ( 165,  566),   0 ),
            Label(["17x87", "DK-11203"],                        ( 17,  87), FormFactor.DIE_CUT,       ( 201, 1026), ( 165,  956),   0 ),
            Label(["23x23", "DK-11221"],                        ( 23,  23), FormFactor.DIE_CUT,       ( 272,  272), ( 202,  202),  42 ),
            Label(["29x42"],                                    ( 29,  42), FormFactor.DIE_CUT,       ( 342,  495), ( 306,  425),   6 ),
            Label(["29x90", "DK-11201"],                        ( 29,  90), FormFactor.DIE_CUT,       ( 342, 1061), ( 306,  991),   6 ),
            Label(["39x90", "DK-11208"],                        ( 38,  90), FormFactor.DIE_CUT,       ( 449, 1061), ( 413,  991),  12 ),
            Label(["39x48"],                                    ( 39,  48), FormFactor.DIE_CUT,       ( 461,  565), ( 425,  495),   6 ),
            Label(["52x29"],                                    ( 52,  29), FormFactor.DIE_CUT,       ( 614,  341), ( 578,  271),   0 ),
            Label(["54x29"],                                    ( 54,  29), FormFactor.DIE_CUT,       ( 630,  341), ( 598,  271),  60 ),
            Label(["60x86", "DK-11234", "DK-12343PK"],          ( 60,  87), FormFactor.DIE_CUT,       ( 708, 1024), ( 672,  954),  18 ),
            Label(["62x29", "DK-11209"],                        ( 62,  29), FormFactor.DIE_CUT,       ( 732,  341), ( 696,  271),  12 ),
            Label(["62x100", "DK-11202"],                       ( 62, 100), FormFactor.DIE_CUT,       ( 732, 1179), ( 696, 1109),  12 ),

            # Round Die-Cut
            Label(["d12", "DK-11219"],                          ( 12,  12), FormFactor.ROUND_DIE_CUT, ( 142,  142), (  94,   94), 113 , feed_margin=35),
            Label(["d24", "DK-11218"],                          ( 24,  24), FormFactor.ROUND_DIE_CUT, ( 284,  284), ( 236,  236),  42 ),
            Label(["d58", "DK-11207"],                          ( 58,  58), FormFactor.ROUND_DIE_CUT, ( 688,  688), ( 618,  618),  51 ),
        ]

class BrotherDeviceQL10(BrotherDeviceQL):
    @property
    def labels(self):
        return super().labels + [
            # Continuous
            Label(["102", "DK-22243"],                          (102,   0), FormFactor.ENDLESS,       (1200,    0), (1164,    0),  12 , feed_margin=35),
            Label(["104"],                                      (104,   0), FormFactor.ENDLESS,       (1227,    0), (1200,    0),  -8 , feed_margin=35),
            
            # Die-Cut
            Label(["102x51", "DK-11240"],                       (102,  51), FormFactor.DIE_CUT,       (1200,  596), (1164,  526),  12),
            Label(["102x152", "DK-11241"],                      (102, 153), FormFactor.DIE_CUT,       (1200, 1804), (1164, 1660),  12),
        ]

class BrotherDeviceQL11(BrotherDeviceQL10):
    @property
    def labels(self):
        return super().labels + [
            # Continuous
            Label(["103", "DK-22246"],                          (104,   0), FormFactor.ENDLESS,       (1224,    0), (1200,    0),  12 , feed_margin=35),
            
            # Die-Cut
            Label(["103x164", "DK-11247"],                      (104, 164), FormFactor.DIE_CUT,       (1224, 1941), (1200, 1822),  12),
        ]

class BrotherDevicePT(BrotherDevice):
    @property
    def labels(self):
        return super().labels + [
            # Continuous
            Label(["12", "pt12"],                               ( 12,   0), FormFactor.PTOUCH_ENDLESS,( 170,    0), ( 150,    0),   213, feed_margin=14),
            Label(["18", "pt18"],                               ( 18,   0), FormFactor.PTOUCH_ENDLESS,( 256,    0), ( 234,    0),   171, feed_margin=14),
            Label(["24", "pt24"],                               ( 24,   0), FormFactor.PTOUCH_ENDLESS,( 128,    0), ( 128,    0),   0, feed_margin=14),
            Label(["36", "pt36"],                               ( 36,   0), FormFactor.PTOUCH_ENDLESS,( 512,    0), ( 454,    0),   61, feed_margin=14),
        ]

class BrotherDevicePTE(BrotherDevice):
    @property
    def labels(self):
        return super().labels + [
            # Continuous
            Label(["6", "pte6"],                                (  6,   0), FormFactor.PTOUCH_ENDLESS,(  42,    0), (  32,    0),  48, feed_margin=14),
            Label(["9", "pte9"],                                (  9,   0), FormFactor.PTOUCH_ENDLESS,(  64,    0), (  50,    0),  39, feed_margin=14),
            Label(["12", "pte12"],                              ( 12,   0), FormFactor.PTOUCH_ENDLESS,(  84,    0), (  70,    0),  29, feed_margin=14),
            Label(["18", "pte18"],                              ( 18,   0), FormFactor.PTOUCH_ENDLESS,( 128,    0), ( 112,    0),   8, feed_margin=14),
            Label(["24", "pte24"],                              ( 24,   0), FormFactor.PTOUCH_ENDLESS,( 170,    0), ( 128,    0),   0, feed_margin=14),
        ]

class BrotherDeviceManager(ElementManager):
    def __init__(self):
        super().__init__((
            # QL Series
            BrotherDeviceQL('QL-500',    (295, 11811), compression=False, mode_setting=False, expanded_mode=False, cutting=False),
            BrotherDeviceQL('QL-550',    (295, 11811), compression=False, mode_setting=False),
            BrotherDeviceQL('QL-560',    (295, 11811), compression=False, mode_setting=False),
            BrotherDeviceQL('QL-570',    (150, 11811), compression=False, mode_setting=False),
            BrotherDeviceQL('QL-580N',   (150, 11811)),
            BrotherDeviceQL('QL-600',    (150, 11811)),
            BrotherDeviceQL('QL-650TD',  (295, 11811)),
            BrotherDeviceQL('QL-700',    (150, 11811), compression=False, mode_setting=False),
            BrotherDeviceQL('QL-710W',   (150, 11811)),
            BrotherDeviceQL('QL-720NW',  (150, 11811)),
            BrotherDeviceQL('QL-800',    (150, 11811), two_color=True, compression=False, num_invalidate_bytes=400),
            BrotherDeviceQL('QL-810W',   (150, 11811), two_color=True, num_invalidate_bytes=400),
            BrotherDeviceQL('QL-820NWB', (150, 11811), two_color=True, num_invalidate_bytes=400),

            # QL 10 Series
            BrotherDeviceQL10('QL-1050',   (295, 35433), number_bytes_per_row=162, additional_offset_r=44),
            BrotherDeviceQL10('QL-1060N',  (295, 35433), number_bytes_per_row=162, additional_offset_r=44),
            
            # QL 11 Series
            BrotherDeviceQL11('QL-1100',   (301, 35434), number_bytes_per_row=162, additional_offset_r=44),
            BrotherDeviceQL11('QL-1100NWB',(301, 35434), number_bytes_per_row=162, additional_offset_r=44),
            BrotherDeviceQL11('QL-1115NWB',(301, 35434), number_bytes_per_row=162, additional_offset_r=44),
            
            # PT Series
            BrotherDevicePT('PT-P750W',  (31, 14172), number_bytes_per_row=16),
            BrotherDevicePT('PT-P900W',  (57, 28346), number_bytes_per_row=70),
            BrotherDevicePT('PT-P950NW', (57, 28346), number_bytes_per_row=70),

            # PTE Series
            BrotherDevicePTE('PT-E550W',  (31, 14172), number_bytes_per_row=16),
        ))
